/*
    webmail -- simple pop3/smtp web e-mail cgi
    Copyright (C) 2000-2003 Dan Cahill

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "main.h"

#define RFC1123FMT "%a, %d %b %Y %H:%M:%S GMT"

void send_header(int cacheable, char *extra_header, char *mime_type, int length, time_t mod)
{
	char timebuf[100];
	static short int sent=0;
	time_t t=0;

	if (sent) return;
	if (length>=0) {
		printf("Content-Length: %d\r\n", length);
	}
	if (mod!=(time_t)-1) {
		strftime(timebuf, sizeof(timebuf), RFC1123FMT, gmtime(&mod));
		printf("Last-Modified: %s\r\n", timebuf);
	}
	if (cacheable) {
		printf("Cache-Control: public\r\n");
		printf("Pragma: public\r\n");
	} else {
		strftime(timebuf, sizeof(timebuf), RFC1123FMT, gmtime(&t));
		printf("Expires: %s\r\n", timebuf);
		printf("Cache-Control: no-cache; must-revalidate\r\n");
		printf("Pragma: no-cache\r\n");
	}
	if (extra_header!=(char*)0) {
		printf("Content-Type: %s\r\n", mime_type);
	} else {
		printf("Content-Type: text/html\r\n");
	}
	printf("\r\n");
	sent=1;
	return;
}

/* see http://www.netscape.com/newsref/std/cookie_spec.html for cookie handling stuff */
int wmcookieget()
{
	char timebuffer[100];
	char temppasswd[100];
	char *ptemp;
	time_t t;

	memset(temppasswd, 0, sizeof(temppasswd));
	memset(wmusername, 0, sizeof(wmusername));
	memset(wmpassword, 0, sizeof(wmpassword));
	memset(wmservertype, 0, sizeof(wmservertype));
	memset(wmpop3server, 0, sizeof(wmpop3server));
	memset(wmsmtpserver, 0, sizeof(wmsmtpserver));
	ptemp=strcasestr(request.Cookie, "wminfo=");
	if (ptemp==NULL) return -1;
	ptemp+=7;
	while ((*ptemp)&&(*ptemp!=':')&&(*ptemp!=';')&&(strlen(wmusername)<sizeof(wmusername)))
		wmusername[strlen(wmusername)]=*ptemp++;
	ptemp++;
	while ((*ptemp)&&(*ptemp!=':')&&(*ptemp!=';')&&(strlen(wmservertype)<sizeof(wmservertype)-1))
		wmservertype[strlen(wmservertype)]=*ptemp++;
	if (config.support_imap!='y') {
		snprintf(wmservertype, sizeof(wmservertype)-1, "POP3");
	}
	if (config.support_pop3!='y') {
		snprintf(wmservertype, sizeof(wmservertype)-1, "IMAP");
	}
	if (config.allow_userhosts=='y') {
		ptemp++;
		while ((*ptemp)&&(*ptemp!=':')&&(*ptemp!=';')&&(strlen(wmpop3server)<sizeof(wmpop3server)-1))
			wmpop3server[strlen(wmpop3server)]=*ptemp++;
		ptemp++;
		while ((*ptemp)&&(*ptemp!=':')&&(*ptemp!=';')&&(strlen(wmsmtpserver)<sizeof(wmsmtpserver)-1))
			wmsmtpserver[strlen(wmsmtpserver)]=*ptemp++;
	} else {
		snprintf(wmpop3server, sizeof(wmpop3server)-1, "%s", config.pop3_host);
		snprintf(wmsmtpserver, sizeof(wmsmtpserver)-1, "%s", config.smtp_host);
	}
	ptemp=strcasestr(request.Cookie, "wmpass=");
	if (ptemp==NULL) return -1;
	ptemp+=7;
	while ((*ptemp)&&(*ptemp!=';')&&(strlen(temppasswd)<sizeof(temppasswd)))
		temppasswd[strlen(temppasswd)]=*ptemp++;
	snprintf(wmpassword, sizeof(wmpassword), "%s", DecodeBase64string(temppasswd));
	t=time((time_t*)0)+604800;
	strftime(timebuffer, sizeof(timebuffer), RFC1123FMT, gmtime(&t));
	printf("Set-Cookie: wminfo=%s:%s:%s:%s; path=%s/; expires=%s\r\n", wmusername, wmservertype, wmpop3server, wmsmtpserver, request.ScriptName, timebuffer);
	printf("Set-Cookie: wmpass=%s; path=%s/\r\n", temppasswd, request.ScriptName);
	if (strlen(wmpassword)==0) return -1;
	return 0;
}

int wmcookieset()
{
	time_t t;
	char timebuffer[100];

	if (strcasecmp(request.RequestMethod, "POST")!=0) return -1;
	if (getpostenv("WMUSERNAME")!=NULL) {
		strncpy(wmusername, getpostenv("WMUSERNAME"), sizeof(wmusername)-1);
	}
	if (getpostenv("WMPASSWORD")!=NULL) {
		strncpy(wmpassword, getpostenv("WMPASSWORD"), sizeof(wmpassword)-1);
	}
	if (getpostenv("WMSERVERTYPE")!=NULL) {
		strncpy(wmservertype, getpostenv("WMSERVERTYPE"), sizeof(wmservertype)-1);
	}
	if (config.support_imap!='y') {
		snprintf(wmservertype, sizeof(wmservertype)-1, "POP3");
	}
	if (config.support_pop3!='y') {
		snprintf(wmservertype, sizeof(wmservertype)-1, "IMAP");
	}
	if (config.allow_userhosts=='y') {
		if (getpostenv("WMPOP3SERVER")!=NULL) {
			strncpy(wmpop3server, getpostenv("WMPOP3SERVER"), sizeof(wmpop3server)-1);
		}
		if (getpostenv("WMSMTPSERVER")!=NULL) {
			strncpy(wmsmtpserver, getpostenv("WMSMTPSERVER"), sizeof(wmsmtpserver)-1);
		}
	} else {
		snprintf(wmpop3server, sizeof(wmpop3server)-1, "%s", config.pop3_host);
		snprintf(wmsmtpserver, sizeof(wmsmtpserver)-1, "%s", config.smtp_host);
	}
	/* thanks to jberkes@pc-tools.net for the syslog addition */
	if (wmserver_connect()==0) {
		t=time((time_t*)0)+604800;
		strftime(timebuffer, sizeof(timebuffer), RFC1123FMT, gmtime(&t));
		printf("Set-Cookie: wminfo=%s:%s:%s:%s; path=%s/; expires=%s\r\n", wmusername, wmservertype, wmpop3server, wmsmtpserver, request.ScriptName, timebuffer);
		printf("Set-Cookie: wmpass=%s; path=%s/\r\n", EncodeBase64string(wmpassword), request.ScriptName);
#ifndef WIN32
		if (config.use_syslog=='y') {
			syslog(LOG_MAIL|LOG_INFO, "%s - %s @ %s success", request.ClientIP, wmusername, wmpop3server);
		}
#endif
		return 0;
	}
#ifndef WIN32
	if (config.use_syslog=='y') {
		syslog(LOG_MAIL|LOG_INFO, "%s - Invalid username or password for %s @ %s", request.ClientIP, wmusername, wmpop3server);
	}
#endif
	wmexit();
	return -1;
}

void wmcookiekill()
{
	printf("Set-Cookie: wmpass=; path=%s/\r\n", request.ScriptName);
	return;
}
